@@ -121,11 +121,11 @@ package
 			want:          nil,
 		},
 		{
-			name:          "package completion after keyword 'package'",
+			name:          "package completion after package keyword",
 			filename:      "fruits/testfile2.go",
 			triggerRegexp: "package()",
 			want:          []string{"package apple", "package apple_test", "package fruits", "package fruits_test", "package main"},
-			editRegexp:    "package\n",
+			editRegexp:    "package",
 		},
 		{
 			name:          "package completion with 'pac' prefix",
@@ -164,14 +164,14 @@ package
 			filename:      "123f_r.u~its-123/testfile.go",
 			triggerRegexp: "package()",
 			want:          []string{"package fruits123", "package fruits123_test", "package main"},
-			editRegexp:    "package\n",
+			editRegexp:    "package",
 		},
 		{
 			name:          "package completion for invalid dir name",
 			filename:      ".invalid-dir@-name/testfile.go",
 			triggerRegexp: "package()",
 			want:          []string{"package main"},
-			editRegexp:    "package\n",
+			editRegexp:    "package",
 		},
 	} {
 		t.Run(tc.name, func(t *testing.T) {
@@ -187,21 +187,30 @@ package
 				// of the file. {Start,End}.Line are zero-based.
 				lineCount := len(strings.Split(env.BufferText(tc.filename), "\n"))
 				for _, item := range completions.Items {
-					if start := int(item.TextEdit.Range.Start.Line); start > lineCount {
-						t.Fatalf("unexpected text edit range start line number: got %d, want <= %d", start, lineCount)
-					}
-					if end := int(item.TextEdit.Range.End.Line); end > lineCount {
-						t.Fatalf("unexpected text edit range end line number: got %d, want <= %d", end, lineCount)
+					for _, mode := range []string{"replace", "insert"} {
+						edit, err := protocol.SelectCompletionTextEdit(item, mode == "replace")
+						if err != nil {
+							t.Fatalf("unexpected text edit in completion item (%v): %v", mode, err)
+						}
+						if start := int(edit.Range.Start.Line); start > lineCount {
+							t.Fatalf("unexpected text edit range (%v) start line number: got %d, want <= %d", mode, start, lineCount)
+						}
+						if end := int(edit.Range.End.Line); end > lineCount {
+							t.Fatalf("unexpected text edit range (%v) end line number: got %d, want <= %d", mode, end, lineCount)
+						}
 					}
 				}
 
 				if tc.want != nil {
 					expectedLoc := env.RegexpSearch(tc.filename, tc.editRegexp)
 					for _, item := range completions.Items {
-						gotRng := item.TextEdit.Range
-						if expectedLoc.Range != gotRng {
-							t.Errorf("unexpected completion range for completion item %s: got %v, want %v",
-								item.Label, gotRng, expectedLoc.Range)
+						for _, mode := range []string{"replace", "insert"} {
+							edit, _ := protocol.SelectCompletionTextEdit(item, mode == "replace")
+							gotRng := edit.Range
+							if expectedLoc.Range != gotRng {
+								t.Errorf("unexpected completion range (%v) for completion item %s: got %v, want %v",
+									mode, item.Label, gotRng, expectedLoc.Range)
+							}
 						}
 					}
 				}
@@ -267,9 +276,6 @@ module mod.com
 go 1.14
 
 require example.com v1.2.3
--- go.sum --
-example.com v1.2.3 h1:ihBTGWGjTU3V4ZJ9OmHITkU9WQ4lGdQkMjgyLFk0FaY=
-example.com v1.2.3/go.mod h1:Y2Rc5rVWjWur0h3pd9aEvK5Pof8YKDANh9gHA2Maujo=
 -- main.go --
 package main
 
@@ -286,6 +292,7 @@ func _() {
 }
 `
 	WithOptions(
+		WriteGoSum("."),
 		ProxyFiles(proxy),
 	).Run(t, mod, func(t *testing.T, env *Env) {
 		// Make sure the dependency is in the module cache and accessible for
@@ -338,9 +345,6 @@ module mod.com
 go 1.14
 
 require example.com v1.2.3
--- go.sum --
-example.com v1.2.3 h1:ihBTGWGjTU3V4ZJ9OmHITkU9WQ4lGdQkMjgyLFk0FaY=
-example.com v1.2.3/go.mod h1:Y2Rc5rVWjWur0h3pd9aEvK5Pof8YKDANh9gHA2Maujo=
 -- useblah.go --
 // +build hidden
 
@@ -352,7 +356,9 @@ package mainmod
 
 const Name = "mainmod"
 `
-	WithOptions(ProxyFiles(proxy)).Run(t, files, func(t *testing.T, env *Env) {
+	WithOptions(
+		WriteGoSum("."),
+		ProxyFiles(proxy)).Run(t, files, func(t *testing.T, env *Env) {
 		env.CreateBuffer("import.go", "package pkg\nvar _ = mainmod.Name\n")
 		env.SaveBuffer("import.go")
 		content := env.ReadWorkspaceFile("import.go")
@@ -462,12 +468,12 @@ module test.com
 go 1.16
 -- prog.go --
 package waste
-// Deprecated, use newFoof
+// Deprecated: use newFoof.
 func fooFunc() bool {
 	return false
 }
 
-// Deprecated
+// Deprecated: bad.
 const badPi = 3.14
 
 func doit() {
@@ -540,6 +546,98 @@ func main() {
 	})
 }
 
+func TestUnimportedCompletion_VSCodeIssue3365(t *testing.T) {
+	const src = `
+-- go.mod --
+module mod.com
+
+go 1.19
+
+-- main.go --
+package main
+
+func main() {
+	println(strings.TLower)
+}
+
+var Lower = ""
+`
+	find := func(t *testing.T, completions *protocol.CompletionList, name string) protocol.CompletionItem {
+		t.Helper()
+		if completions == nil || len(completions.Items) == 0 {
+			t.Fatalf("no completion items")
+		}
+		for _, i := range completions.Items {
+			if i.Label == name {
+				return i
+			}
+		}
+		t.Fatalf("no item with label %q", name)
+		return protocol.CompletionItem{}
+	}
+
+	for _, supportInsertReplace := range []bool{true, false} {
+		t.Run(fmt.Sprintf("insertReplaceSupport=%v", supportInsertReplace), func(t *testing.T) {
+			capabilities := fmt.Sprintf(`{ "textDocument": { "completion": { "completionItem": {"insertReplaceSupport":%t, "snippetSupport": false } } } }`, supportInsertReplace)
+			runner := WithOptions(CapabilitiesJSON([]byte(capabilities)))
+			runner.Run(t, src, func(t *testing.T, env *Env) {
+				env.OpenFile("main.go")
+				env.Await(env.DoneWithOpen())
+				orig := env.BufferText("main.go")
+
+				// We try to trigger completion at "println(strings.T<>Lower)"
+				// and accept the completion candidate that matches the 'accept' label.
+				insertModeWant := "println(strings.ToUpperLower)"
+				if !supportInsertReplace {
+					insertModeWant = "println(strings.ToUpper)"
+				}
+				testcases := []struct {
+					mode   string
+					accept string
+					want   string
+				}{
+					{
+						mode:   "insert",
+						accept: "ToUpper",
+						want:   insertModeWant,
+					},
+					{
+						mode:   "insert",
+						accept: "ToLower",
+						want:   "println(strings.ToLower)", // The suffix 'Lower' is included in the text edit.
+					},
+					{
+						mode:   "replace",
+						accept: "ToUpper",
+						want:   "println(strings.ToUpper)",
+					},
+					{
+						mode:   "replace",
+						accept: "ToLower",
+						want:   "println(strings.ToLower)",
+					},
+				}
+
+				for _, tc := range testcases {
+					t.Run(fmt.Sprintf("%v/%v", tc.mode, tc.accept), func(t *testing.T) {
+
+						env.SetSuggestionInsertReplaceMode(tc.mode == "replace")
+						env.SetBufferContent("main.go", orig)
+						loc := env.RegexpSearch("main.go", `Lower\)`)
+						completions := env.Completion(loc)
+						item := find(t, completions, tc.accept)
+						env.AcceptCompletion(loc, item)
+						env.Await(env.DoneWithChange())
+						got := env.BufferText("main.go")
+						if !strings.Contains(got, tc.want) {
+							t.Errorf("unexpected state after completion:\n%v\nwanted %v", got, tc.want)
+						}
+					})
+				}
+			})
+		})
+	}
+}
 func TestUnimportedCompletionHasPlaceholders60269(t *testing.T) {
 	// We can't express this as a marker test because it doesn't support AcceptCompletion.
 	const src = `
@@ -869,6 +967,275 @@ use ./missing/
 	})
 }
 
+const reverseInferenceSrcPrelude = `
+-- go.mod --
+module mod.com
+
+go 1.18
+-- a.go --
+package a
+
+type InterfaceA interface {
+	implA()
+}
+
+type InterfaceB interface {
+	implB()
+}
+
+
+type TypeA struct{}
+
+func (TypeA) implA() {}
+
+type TypeX string
+
+func (TypeX) implB() {}
+
+type TypeB struct{}
+
+func (TypeB) implB() {}
+
+type TypeC struct{} // should have no impact
+
+type Wrap[T any] struct {
+	inner *T
+}
+
+func NewWrap[T any](x T) Wrap[T] {
+	return Wrap[T]{inner: &x}
+}
+
+func DoubleWrap[T any, U any](t T, u U) (Wrap[T], Wrap[U]) {
+	return Wrap[T]{inner: &t}, Wrap[U]{inner: &u}
+}
+
+func IntWrap[T int32 | int64](x T) Wrap[T] {
+	return Wrap[T]{inner: &x}
+}
+
+var ia InterfaceA
+var ib InterfaceB
+
+var avar TypeA
+var bvar TypeB
+
+var i int
+var i32 int32
+var i64 int64
+`
+
+func TestReverseInferCompletion(t *testing.T) {
+	src := reverseInferenceSrcPrelude + `
+	func main() {
+		var _ Wrap[int64] = IntWrap()
+	}
+	`
+	Run(t, src, func(t *testing.T, env *Env) {
+		compl := env.RegexpSearch("a.go", `IntWrap\(()\)`)
+
+		env.OpenFile("a.go")
+		result := env.Completion(compl)
+
+		wantLabel := []string{"i64", "i", "i32", "int64()"}
+
+		// only check the prefix due to formatting differences with escaped characters
+		wantText := []string{"i64", "int64(i", "int64(i32", "int64("}
+
+		for i, item := range result.Items[:len(wantLabel)] {
+			if diff := cmp.Diff(wantLabel[i], item.Label); diff != "" {
+				t.Errorf("Completion: unexpected label mismatch (-want +got):\n%s", diff)
+			}
+
+			if insertText, ok := item.TextEdit.Value.(protocol.InsertReplaceEdit); ok {
+				if diff := cmp.Diff(wantText[i], insertText.NewText[:len(wantText[i])]); diff != "" {
+					t.Errorf("Completion: unexpected insertText mismatch (checks prefix only) (-want +got):\n%s", diff)
+				}
+			}
+		}
+	})
+}
+
+func TestInterfaceReverseInferCompletion(t *testing.T) {
+	src := reverseInferenceSrcPrelude + `
+	func main() {
+		var wa Wrap[InterfaceA]
+		var wb Wrap[InterfaceB]
+		wb = NewWrap() // wb is of type Wrap[InterfaceB]
+	}
+	`
+
+	Run(t, src, func(t *testing.T, env *Env) {
+		compl := env.RegexpSearch("a.go", `NewWrap\(()\)`)
+
+		env.OpenFile("a.go")
+		result := env.Completion(compl)
+
+		wantLabel := []string{"ib", "bvar", "wb.inner", "TypeB{}", "TypeX()", "nil"}
+
+		// only check the prefix due to formatting differences with escaped characters
+		wantText := []string{"ib", "InterfaceB(", "*wb.inner", "InterfaceB(", "InterfaceB(", "nil"}
+
+		for i, item := range result.Items[:len(wantLabel)] {
+			if diff := cmp.Diff(wantLabel[i], item.Label); diff != "" {
+				t.Errorf("Completion: unexpected label mismatch (-want +got):\n%s", diff)
+			}
+
+			if insertText, ok := item.TextEdit.Value.(protocol.InsertReplaceEdit); ok {
+				if diff := cmp.Diff(wantText[i], insertText.NewText[:len(wantText[i])]); diff != "" {
+					t.Errorf("Completion: unexpected insertText mismatch (checks prefix only) (-want +got):\n%s", diff)
+				}
+			}
+		}
+	})
+}
+
+func TestInvalidReverseInferenceDefaultsToConstraintCompletion(t *testing.T) {
+	src := reverseInferenceSrcPrelude + `
+	func main() {
+		var wa Wrap[InterfaceA]
+		// This is ambiguous, so default to the constraint rather the inference.
+		wa = IntWrap()
+	}
+	`
+	Run(t, src, func(t *testing.T, env *Env) {
+		compl := env.RegexpSearch("a.go", `IntWrap\(()\)`)
+
+		env.OpenFile("a.go")
+		result := env.Completion(compl)
+
+		wantLabel := []string{"i32", "i64", "nil"}
+
+		for i, item := range result.Items[:len(wantLabel)] {
+			if diff := cmp.Diff(wantLabel[i], item.Label); diff != "" {
+				t.Errorf("Completion: unexpected label mismatch (-want +got):\n%s", diff)
+			}
+		}
+	})
+}
+
+func TestInterfaceReverseInferTypeParamCompletion(t *testing.T) {
+	src := reverseInferenceSrcPrelude + `
+	func main() {
+		var wa Wrap[InterfaceA]
+		var wb Wrap[InterfaceB]
+		wb = NewWrap[]()
+	}
+	`
+
+	Run(t, src, func(t *testing.T, env *Env) {
+		compl := env.RegexpSearch("a.go", `NewWrap\[()\]\(\)`)
+
+		env.OpenFile("a.go")
+		result := env.Completion(compl)
+		want := []string{"InterfaceB", "TypeB", "TypeX", "InterfaceA", "TypeA"}
+		for i, item := range result.Items[:len(want)] {
+			if diff := cmp.Diff(want[i], item.Label); diff != "" {
+				t.Errorf("Completion: unexpected mismatch (-want +got):\n%s", diff)
+			}
+		}
+	})
+}
+
+func TestInvalidReverseInferenceTypeParamDefaultsToConstraintCompletion(t *testing.T) {
+	src := reverseInferenceSrcPrelude + `
+	func main() {
+		var wa Wrap[InterfaceA]
+		// This is ambiguous, so default to the constraint rather the inference.
+		wb = IntWrap[]()
+	}
+	`
+
+	Run(t, src, func(t *testing.T, env *Env) {
+		compl := env.RegexpSearch("a.go", `IntWrap\[()\]\(\)`)
+
+		env.OpenFile("a.go")
+		result := env.Completion(compl)
+		want := []string{"int32", "int64"}
+		for i, item := range result.Items[:len(want)] {
+			if diff := cmp.Diff(want[i], item.Label); diff != "" {
+				t.Errorf("Completion: unexpected mismatch (-want +got):\n%s", diff)
+			}
+		}
+	})
+}
+
+func TestReverseInferDoubleTypeParamCompletion(t *testing.T) {
+	src := reverseInferenceSrcPrelude + `
+	func main() {
+		var wa Wrap[InterfaceA]
+		var wb Wrap[InterfaceB]
+
+		wa, wb = DoubleWrap[]()
+		// _ is necessary to trick the parser into an index list expression
+		wa, wb = DoubleWrap[InterfaceA, _]()
+	}
+	`
+	Run(t, src, func(t *testing.T, env *Env) {
+		env.OpenFile("a.go")
+
+		compl := env.RegexpSearch("a.go", `DoubleWrap\[()\]\(\)`)
+		result := env.Completion(compl)
+
+		wantLabel := []string{"InterfaceA", "TypeA", "InterfaceB", "TypeB", "TypeC"}
+
+		for i, item := range result.Items[:len(wantLabel)] {
+			if diff := cmp.Diff(wantLabel[i], item.Label); diff != "" {
+				t.Errorf("Completion: unexpected label mismatch (-want +got):\n%s", diff)
+			}
+		}
+
+		compl = env.RegexpSearch("a.go", `DoubleWrap\[InterfaceA, (_)\]\(\)`)
+		result = env.Completion(compl)
+
+		wantLabel = []string{"InterfaceB", "TypeB", "TypeX", "InterfaceA", "TypeA"}
+
+		for i, item := range result.Items[:len(wantLabel)] {
+			if diff := cmp.Diff(wantLabel[i], item.Label); diff != "" {
+				t.Errorf("Completion: unexpected label mismatch (-want +got):\n%s", diff)
+			}
+		}
+	})
+}
+
+func TestDoubleParamReturnCompletion(t *testing.T) {
+	src := reverseInferenceSrcPrelude + `
+	func concrete() (Wrap[InterfaceA], Wrap[InterfaceB]) {
+		return DoubleWrap[]()
+	}
+
+	func concrete2() (Wrap[InterfaceA], Wrap[InterfaceB]) {
+		return DoubleWrap[InterfaceA, _]()
+	}
+	`
+
+	Run(t, src, func(t *testing.T, env *Env) {
+		env.OpenFile("a.go")
+
+		compl := env.RegexpSearch("a.go", `DoubleWrap\[()\]\(\)`)
+		result := env.Completion(compl)
+
+		wantLabel := []string{"InterfaceA", "TypeA", "InterfaceB", "TypeB", "TypeC"}
+
+		for i, item := range result.Items[:len(wantLabel)] {
+			if diff := cmp.Diff(wantLabel[i], item.Label); diff != "" {
+				t.Errorf("Completion: unexpected label mismatch (-want +got):\n%s", diff)
+			}
+		}
+
+		compl = env.RegexpSearch("a.go", `DoubleWrap\[InterfaceA, (_)\]\(\)`)
+		result = env.Completion(compl)
+
+		wantLabel = []string{"InterfaceB", "TypeB", "TypeX", "InterfaceA", "TypeA"}
+
+		for i, item := range result.Items[:len(wantLabel)] {
+			if diff := cmp.Diff(wantLabel[i], item.Label); diff != "" {
+				t.Errorf("Completion: unexpected label mismatch (-want +got):\n%s", diff)
+			}
+		}
+	})
+}
+
 func TestBuiltinCompletion(t *testing.T) {
 	const files = `
 -- go.mod --
